"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import csv
import io
from concurrent.futures import Future
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Set, Union

import agate
import dbt.exceptions
from dbt.adapters.base import AdapterConfig, available
from dbt.adapters.base.impl import catch_as_completed
from dbt.adapters.base.relation import BaseRelation, InformationSchema
from dbt.adapters.sql import SQLAdapter
from dbt.clients.agate_helper import table_from_rows
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.relation import RelationType
from dbt.events import AdapterLogger
from dbt.utils import executor, filter_null_values

from dbt.adapters.bytehouse.column import ByteHouseColumn
from dbt.adapters.bytehouse.connections import ByteHouseConnectionManager
from dbt.adapters.bytehouse.relation import ByteHouseRelation

GET_CATALOG_MACRO_NAME = 'get_catalog'
LIST_SCHEMAS_MACRO_NAME = 'list_schemas'


@dataclass
class ByteHouseConfig(AdapterConfig):
    engine: str = 'MergeTree()'
    order_by: Optional[Union[List[str], str]] = 'tuple()'
    partition_by: Optional[Union[List[str], str]] = None


class ByteHouseAdapter(SQLAdapter):
    Relation = ByteHouseRelation
    Column = ByteHouseColumn
    ConnectionManager = ByteHouseConnectionManager
    AdapterSpecificConfigs = ByteHouseConfig
    logger = AdapterLogger("dbt_bytehouse_tests")

    @classmethod
    def date_function(cls):
        return 'now()'

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return 'String'

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))
        # We match these type to the Column.TYPE_LABELS for consistency
        return 'Float32' if decimals else 'Int32'

    @classmethod
    def convert_boolean_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return 'Int32'

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return 'DateTime'

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return 'Date'

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        raise dbt.exceptions.NotImplementedException(
            '`convert_time_type` is not implemented for this adapter!'
        )

    @available.parse(lambda *a, **k: {})
    def get_bytehouse_cluster_name(self):
        conn = self.connections.get_if_exists()
        if conn.credentials.cluster:
            return '"{}"'.format(conn.credentials.cluster)

    @available
    def bytehouse_db_engine_clause(self):
        conn = self.connections.get_if_exists()
        if conn and conn.credentials.database_engine:
            return f'ENGINE {conn.credentials.database_engine}'
        return ''

    @available
    def is_before_version(self, version: str) -> bool:
        conn = self.connections.get_if_exists()
        if conn:
            server_version = conn.handle.server_version
            return compare_versions(version, server_version) > 0
        return False

    @available
    def supports_atomic_exchange(self) -> bool:
        conn = self.connections.get_if_exists()
        return conn and conn.handle.atomic_exchange

    @available
    def can_exchange(self, schema: str, rel_type: str) -> bool:
        if rel_type != 'table' or not schema or not self.supports_atomic_exchange():
            return False
        ch_db = self.get_ch_database(schema)
        return ch_db and ch_db.engine in ('Atomic', 'Replicated')

    def check_schema_exists(self, database, schema):
        results = self.execute_macro(LIST_SCHEMAS_MACRO_NAME, kwargs={'database': database})
        return schema in (row[0] for row in results)

    def drop_schema(self, relation: BaseRelation) -> None:
        super().drop_schema(relation)
        conn = self.connections.get_if_exists()
        if conn:
            conn.handle.database_dropped(relation.schema)

    def _make_match_kwargs(self, database: str, schema: str, identifier: str) -> Dict[str, str]:
        return filter_null_values(
            {
                "database": database,
                "identifier": identifier,
            }
        )

    def list_relations_without_caching(
        self, schema_relation: ByteHouseRelation
    ) -> List[ByteHouseRelation]:
        kwargs = {'schema_relation': schema_relation}
        results = self.execute_macro('list_relations_without_caching', kwargs=kwargs)
        conn_supports_exchange = self.supports_atomic_exchange()

        relations = []
        for row in results:
            name, schema, type_info, db_engine = row
            rel_type = RelationType.View if 'view' in type_info else RelationType.Table
            can_exchange = (
                conn_supports_exchange
                and rel_type == RelationType.Table
                and db_engine in ('Atomic', 'Replicated')
            )
            relation = self.Relation.create(
                database=None,
                schema=schema,
                identifier=name,
                type=rel_type,
                can_exchange=can_exchange,
            )
            relations.append(relation)

        return relations

    def get_relation(self, database: Optional[str], schema: str, identifier: str):
        if not self.Relation.include_policy.database:
            database = None

        return super().get_relation(database, schema, identifier)

    @available
    def get_ch_database(self, schema: str):
        try:
            results = self.execute_macro('bytehouse__get_database', kwargs={'database': schema})
            if len(results.rows):
                return ByteHouseDatabase(**results.rows[0])
            return None
        except dbt.exceptions.RuntimeException:
            return None

    def parse_bytehouse_columns(
        self, relation: ByteHouseRelation, raw_rows: List[agate.Row]
    ) -> List[ByteHouseColumn]:
        rows = [dict(zip(row._keys, row._values)) for row in raw_rows]

        return [
            ByteHouseColumn(
                column=column['name'],
                dtype=column['type'],
            )
            for column in rows
        ]

    def get_columns_in_relation(self, relation: ByteHouseRelation) -> List[ByteHouseColumn]:
        rows: List[agate.Row] = super().get_columns_in_relation(relation)
        return self.parse_bytehouse_columns(relation, rows)

    def get_catalog(self, manifest):
        schema_map = self._get_catalog_schemas(manifest)

        with executor(self.config) as tpe:
            futures: List[Future[agate.Table]] = []
            for info, schemas in schema_map.items():
                for schema in schemas:
                    futures.append(
                        tpe.submit_connected(
                            self,
                            schema,
                            self._get_one_catalog,
                            info,
                            [schema],
                            manifest,
                        )
                    )
            catalogs, exceptions = catch_as_completed(futures)
        return catalogs, exceptions

    def _get_one_catalog(
        self,
        information_schema: InformationSchema,
        schemas: Set[str],
        manifest: Manifest,
    ) -> agate.Table:
        if len(schemas) != 1:
            dbt.exceptions.raise_compiler_error(
                f'Expected only one schema in bytehouse _get_one_catalog, found ' f'{schemas}'
            )

        return super()._get_one_catalog(information_schema, schemas, manifest)

    @classmethod
    def _catalog_filter_table(cls, table: agate.Table, manifest: Manifest) -> agate.Table:
        table = table_from_rows(
            table.rows,
            table.column_names,
            text_only_columns=['table_schema', 'table_name'],
        )
        return table.where(_catalog_filter_schemas(manifest))

    def get_rows_different_sql(
        self,
        relation_a: ByteHouseRelation,
        relation_b: ByteHouseRelation,
        column_names: Optional[List[str]] = None,
        except_operator: str = None,
    ) -> str:
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((self.quote(c.name) for c in columns))
        else:
            names = sorted((self.quote(n) for n in column_names))

        alias_a = 'ta'
        alias_b = 'tb'
        columns_csv_a = ', '.join([f'{alias_a}.{name}' for name in names])
        columns_csv_b = ', '.join([f'{alias_b}.{name}' for name in names])
        join_condition = ' AND '.join([f'{alias_a}.{name} = {alias_b}.{name}' for name in names])
        first_column = names[0]

        # ByteHouse doesn't have an EXCEPT operator
        sql = COLUMNS_EQUAL_SQL.format(
            alias_a=alias_a,
            alias_b=alias_b,
            first_column=first_column,
            columns_a=columns_csv_a,
            columns_b=columns_csv_b,
            join_condition=join_condition,
            relation_a=str(relation_a),
            relation_b=str(relation_b),
        )

        return sql

    def update_column_sql(
        self,
        dst_name: str,
        dst_column: str,
        clause: str,
        where_clause: Optional[str] = None,
    ) -> str:
        clause = f'alter table {dst_name} update {dst_column} = {clause}'
        if where_clause is not None:
            clause += f' where {where_clause}'
        return clause

    @available
    def get_csv_data(self, table):
        # TODO: Consider passing table to macro
        row_count = len(table.rows)
        column_count = 0
        if row_count > 0:
            column_count = len(table.rows[0])

        # TODO: Use hashmap for conversion
        # TODO: Implement all datatypes
        datatypes = []
        for c in table._column_types:
            if isinstance(c, dbt.clients.agate_helper.Number):
                datatypes.append("Int")
            elif isinstance(c, dbt.clients.agate_helper.ISODateTime):
                datatypes.append("Datetime")
            elif isinstance(c, agate.data_types.boolean.Boolean):
                datatypes.append("BooleanInt")
            elif isinstance(c, agate.data_types.date_time.DateTime):
                datatypes.append("Datetime")
            elif isinstance(c, agate.data_types.number.Number):
                datatypes.append("Int")
            elif isinstance(c, agate.data_types.text.Text):
                datatypes.append("String")
            else:
                datatypes.append("NotImplemented")

        data = "|" + str(row_count)
        data += "|" + str(column_count)
        for datatype in datatypes:
            data += "|" + datatype
        for row in table.rows:
            for datatype, val in enumerate(row):
                data += "|" + str(val)
        return data

    def run_sql_for_tests(self, sql, fetch, conn):
        client = conn.handle
        try:
            if fetch:
                result = client.query(sql).result_set
            else:
                result = client.command(sql)
            if fetch == "one" and len(result) > 0:
                return result[0]
            if fetch == "all":
                return result
        except BaseException as e:
            self.logger.error(sql)
            self.logger.error(e)
            raise
        finally:
            conn.state = 'close'

    @available
    def get_model_settings(self, model):
        settings = model['config'].get('settings', dict())
        res = []
        for key in settings:
            res.append(f' {key}={settings[key]}')
        return '' if len(res) == 0 else 'SETTINGS ' + ', '.join(res) + '\n'


@dataclass
class ByteHouseDatabase:
    name: str
    engine: str
    comment: str


def _expect_row_value(key: str, row: agate.Row):
    if key not in row.keys():
        raise dbt.exceptions.InternalException(
            f'Got a row without \'{key}\' column, columns: {row.keys()}'
        )

    return row[key]


def _catalog_filter_schemas(manifest: Manifest) -> Callable[[agate.Row], bool]:
    schemas = frozenset((None, s.lower()) for d, s in manifest.get_used_schemas())

    def test(row: agate.Row) -> bool:
        table_database = _expect_row_value('table_database', row)
        table_schema = _expect_row_value('table_schema', row)
        if table_schema is None:
            return False
        return (table_database, table_schema.lower()) in schemas

    return test


def compare_versions(v1: str, v2: str) -> int:
    v1_parts = v1.split('.')
    v2_parts = v2.split('.')
    for part1, part2 in zip(v1_parts, v2_parts):
        try:
            if int(part1) != int(part2):
                return 1 if int(part1) > int(part2) else -1
        except ValueError:
            raise dbt.exceptions.RuntimeException(
                "Version must consist of only numbers separated by '.'"
            )
    # Versions are equal - return False
    return 0


COLUMNS_EQUAL_SQL = '''
SELECT
    row_count_diff.difference as row_count_difference,
    diff_count.num_missing as num_mismatched
FROM (
    SELECT
        1 as id,
        (SELECT COUNT(*) as num_rows FROM {relation_a}) -
        (SELECT COUNT(*) as num_rows FROM {relation_b}) as difference
    ) as row_count_diff
INNER JOIN (
    SELECT
        1 as id,
        COUNT(*) as num_missing FROM (
            SELECT
                {columns_a}
            FROM {relation_a} as {alias_a}
            LEFT OUTER JOIN {relation_b} as {alias_b}
                ON {join_condition}
            WHERE {alias_b}.{first_column} IS NULL
            UNION ALL
            SELECT
                {columns_b}
            FROM {relation_b} as {alias_b}
            LEFT OUTER JOIN {relation_a} as {alias_a}
                ON {join_condition}
            WHERE {alias_a}.{first_column} IS NULL
        ) as missing
    ) as diff_count ON row_count_diff.id = diff_count.id
'''.strip()



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from dbt.adapters.base import AdapterPlugin

from dbt.adapters.bytehouse.column import ByteHouseColumn  # noqa
from dbt.adapters.bytehouse.connections import ByteHouseConnectionManager  # noqa
from dbt.adapters.bytehouse.credentials import ByteHouseCredentials
from dbt.adapters.bytehouse.impl import ByteHouseAdapter
from dbt.adapters.bytehouse.relation import ByteHouseRelation  # noqa
from dbt.include import bytehouse  # noqa

Plugin = AdapterPlugin(
    adapter=ByteHouseAdapter,
    credentials=ByteHouseCredentials,
    include_path=bytehouse.PACKAGE_PATH,
)



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import re
from dataclasses import dataclass
from typing import Any, TypeVar

from dbt.adapters.base.column import Column
from dbt.exceptions import RuntimeException

Self = TypeVar('Self', bound='ByteHouseColumn')


@dataclass
class ByteHouseColumn(Column):
    TYPE_LABELS = {
        'STRING': 'String',
        'TIMESTAMP': 'DateTime',
        'FLOAT': 'Float32',
        'INTEGER': 'Int32',
    }
    is_nullable: bool = False
    _brackets_regex = re.compile(r'^(Nullable|LowCardinality)\((.*)\)$')
    _fix_size_regex = re.compile(r'FixedString\((.*?)\)')
    _decimal_regex = re.compile(r'Decimal\((\d+), (\d+)\)')

    def __init__(self, column: str, dtype: str) -> None:
        char_size = None
        numeric_precision = None
        numeric_scale = None

        inner_dtype = self.match_brackets(dtype)
        if inner_dtype:
            dtype = inner_dtype
            if not self.is_nullable:
                # Support LowCardinality(Nullable(dtype))
                inner_dtype = self.match_brackets(dtype)
                dtype = inner_dtype if inner_dtype else dtype

        if dtype.lower().startswith('fixedstring'):
            match_sized = self._fix_size_regex.search(dtype)
            if match_sized:
                char_size = int(match_sized.group(1))

        if dtype.lower().startswith('decimal'):
            match_dec = self._decimal_regex.search(dtype)
            numeric_precision = 0
            numeric_scale = 0
            if match_dec:
                numeric_precision = int(match_dec.group(1))
                numeric_scale = int(match_dec.group(2))

        super().__init__(column, dtype, char_size, numeric_precision, numeric_scale)

    def __repr__(self) -> str:
        return f'<ByteHouseColumn {self.name} ({self.data_type}, is nullable: {self.is_nullable})>'

    @property
    def data_type(self) -> str:
        if self.is_string():
            data_t = self.string_type(self.string_size())
            if self.is_nullable:
                return "Nullable({})".format(data_t)
            return data_t
        elif self.is_numeric():
            data_t = self.numeric_type(self.dtype, self.numeric_precision, self.numeric_scale)
            if self.is_nullable:
                return "Nullable({})".format(data_t)
            return data_t
        else:
            if self.is_nullable:
                return "Nullable({})".format(self.dtype)
            return self.dtype

    def is_string(self) -> bool:
        return self.dtype.lower() in [
            'string',
            'fixedstring',
            'longblob',
            'longtext',
            'tinytext',
            'text',
            'varchar',
            'mediumblob',
            'blob',
            'tinyblob',
            'char',
            'mediumtext',
        ] or self.dtype.lower().startswith('fixedstring')

    def is_integer(self) -> bool:
        return self.dtype.lower().startswith('int') or self.dtype.lower().startswith('uint')

    def is_numeric(self) -> bool:
        return self.dtype.lower().startswith('decimal')

    def is_float(self) -> bool:
        return self.dtype.lower().startswith('float')

    def string_size(self) -> int:
        if not self.is_string():
            raise RuntimeException('Called string_size() on non-string field!')

        if not self.dtype.lower().startswith('fixedstring') or self.char_size is None:
            return 256
        else:
            return int(self.char_size)

    @classmethod
    def string_type(cls, size: int) -> str:
        return 'String'

    @classmethod
    def numeric_type(cls, dtype: str, precision: Any, scale: Any) -> str:
        return f'Decimal({precision}, {scale})'

    def literal(self, value):
        return f'to{self.dtype}({value})'

    def can_expand_to(self, other_column: 'Column') -> bool:
        if not self.is_string() or not other_column.is_string():
            return False

        return other_column.string_size() > self.string_size()

    def match_brackets(self, dtype):
        match = self._brackets_regex.search(dtype.strip())
        if match:
            self.is_nullable = match.group(1) == 'Nullable'
            return match.group(2)



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from dataclasses import dataclass
from typing import Any, Dict, Optional

from dbt.contracts.connection import Credentials

import dbt


@dataclass
class ByteHouseCredentials(Credentials):
    """
    ByteHouse connection credentials data class.
    """

    driver: Optional[str] = None
    host: str = 'localhost'
    port: Optional[int] = None
    region: Optional[str] = None
    account: Optional[str] = None
    user: Optional[str] = 'default'
    retries: int = 1
    database: Optional[str] = None
    schema: Optional[str] = 'default'
    password: str = ''
    warehouse: Optional[str] = None
    cluster: Optional[str] = None
    database_engine: Optional[str] = None
    cluster_mode: bool = False
    secure: bool = False
    verify: bool = True
    connect_timeout: int = 10
    send_receive_timeout: int = 300
    sync_request_timeout: int = 5
    compress_block_size: int = 1048576
    compression: str = ''
    check_exchange: bool = True
    custom_settings: Optional[Dict[str, Any]] = None

    @property
    def type(self):
        return 'bytehouse'

    @property
    def unique_field(self):
        return self.host

    def __post_init__(self):
        if self.database is not None and self.database != self.schema:
            raise dbt.exceptions.RuntimeException(
                f'    schema: {self.schema} \n'
                f'    database: {self.database} \n'
                f'    cluster: {self.cluster} \n'
                f'On ByteHouse, database must be omitted or have the same value as'
                f' schema.'
            )
        self.database = None

    def _connection_keys(self):
        return (
            'driver',
            'host',
            'port',
            'region',
            'account',
            'user',
            'schema',
            'retries',
            'database_engine',
            'cluster_mode',
            'secure',
            'verify',
            'connect_timeout',
            'send_receive_timeout',
            'sync_request_timeout',
            'compress_block_size',
            'compression',
            'check_exchange',
            'custom_settings',
        )



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from abc import ABC, abstractmethod

from dbt.events import AdapterLogger
from dbt.exceptions import FailedToConnectException

from dbt.adapters.bytehouse.credentials import ByteHouseCredentials

logger = AdapterLogger('bytehouse')


def get_db_client(credentials: ByteHouseCredentials):
    credentials.driver = 'native'
    try:
        import bytehouse_driver  # noqa

        from dbt.adapters.bytehouse.nativeclient import BhNativeClient

        return BhNativeClient(credentials)
    except ImportError:
        raise FailedToConnectException(
            'Native adapter required but package bytehouse-driver is not installed'
        )


class BhRetryableException(Exception):
    pass


class BhClientWrapper(ABC):
    def __init__(self, credentials: ByteHouseCredentials):
        self.database = credentials.schema
        self._conn_settings = credentials.custom_settings or {}
        if credentials.cluster_mode or credentials.database_engine == 'Replicated':
            self._conn_settings['database_replicated_enforce_synchronous_settings'] = '1'
            self._conn_settings['insert_quorum'] = 'auto'
        self._client = self._create_client(credentials)
        check_exchange = credentials.check_exchange and not credentials.cluster_mode
        try:
            self._ensure_database(credentials.database_engine)
            self.server_version = self._server_version()
            self.atomic_exchange = not check_exchange or self._check_atomic_exchange()
        except Exception as ex:
            self.close()
            raise ex

    @abstractmethod
    def query(self, sql: str, **kwargs):
        pass

    @abstractmethod
    def command(self, sql: str, **kwargs):
        pass

    def database_dropped(self, database: str):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def _create_client(self, credentials: ByteHouseCredentials):
        pass

    @abstractmethod
    def _set_client_database(self):
        pass

    @abstractmethod
    def _server_version(self):
        pass

    def _ensure_database(self, database_engine) -> None:
        if not self.database:
            return
        # TODO: log & exception handling
        self.command(f'CREATE DATABASE IF NOT EXISTS {self.database}')
        self._set_client_database()

    def _check_atomic_exchange(self) -> bool:
        return False



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import bytehouse_driver
from dateutil import parser
from bytehouse_driver import Client
from bytehouse_driver.errors import NetworkError, SocketTimeoutError
from dbt.exceptions import DatabaseException as DBTDatabaseException
from dbt.version import __version__ as dbt_version

from dbt.adapters.bytehouse import ByteHouseCredentials
from dbt.adapters.bytehouse.dbclient import BhClientWrapper, BhRetryableException


class BhNativeClient(BhClientWrapper):
    def query(self, sql, **kwargs):
        self.prepare_system_database(sql)
        sql = self.rewrite_sql(sql)
        if "rename table" in sql:
            sql = self.get_rename_view_sql(sql)
        if "drop table" in sql or "DROP TABLE" in sql:
            sql = self.modify_drop_table_syntax(sql)
        try:
            return NativeClientResult(self._client.execute(sql, with_column_types=True, **kwargs))
        except bytehouse_driver.errors.Error as ex:
            raise DBTDatabaseException(str(ex).strip()) from ex

    def command(self, sql, **kwargs):
        self.prepare_system_database(sql)
        sql = self.rewrite_sql(sql)
        if "rename table" in sql:
            sql = self.get_rename_view_sql(sql)
        if "drop table" in sql or "DROP TABLE" in sql:
            sql = self.modify_drop_table_syntax(sql)
        try:
            # TODO: Check whether Open source clickhouse support insertion without data block
            if "ByteHouseCustomCSV" in sql:
                query, data = self.handle_format_csv(sql)
                result = self._client.execute(query, data, with_column_types=True, **kwargs)
                return result
            result = self._client.execute(sql, **kwargs)
            if len(result) and len(result[0]):
                return result[0][0]
        except bytehouse_driver.errors.Error as ex:
            raise DBTDatabaseException(str(ex).strip()) from ex

    def close(self):
        self._client.disconnect()

    def _create_client(self, credentials: ByteHouseCredentials):
        client = bytehouse_driver.Client(
            host=credentials.host,
            port=credentials.port,
            region=credentials.region,
            account=credentials.account,
            user=credentials.user,
            password=credentials.password,
            client_name=f'dbt-{dbt_version}',
            secure=credentials.secure,
            verify=credentials.verify,
            connect_timeout=credentials.connect_timeout,
            send_receive_timeout=credentials.send_receive_timeout,
            sync_request_timeout=credentials.sync_request_timeout,
            compress_block_size=credentials.compress_block_size,
            compression=False if credentials.compression == '' else credentials.compression,
            settings=self._conn_settings,
        )
        try:
            client.connection.connect()
            if credentials.warehouse is not None and credentials.warehouse != "":
                if self._is_warehouse_up(client, credentials.warehouse) is False:
                    client.execute('resume warehouse {}'.format(credentials.warehouse))
                client.execute('set warehouse {}'.format(credentials.warehouse))
        except (SocketTimeoutError, NetworkError) as ex:
            raise BhRetryableException(str(ex)) from ex
        return client

    def _is_warehouse_up(self, client: Client, vw_name):
        warehouses = client.execute("show warehouses")
        for warehouse in warehouses:
            if warehouse[1] == vw_name and warehouse[6] == "up":
                return True
        return False

    def _set_client_database(self):
        # After we know the database exists, reconnect to that database if appropriate
        if self._client.connection.database != self.database:
            self._client.connection.disconnect()
            self._client.connection.database = self.database
            self._client.connection.connect()

    def _server_version(self):
        server_info = self._client.connection.server_info
        return (
            f'{server_info.version_major}.{server_info.version_minor}.{server_info.version_patch}'
        )

    def prepare_system_database(self, sql):
        if "system_meta" not in sql:
            return
        # TODO: Add log & exception handling
        self._client.execute("DROP DATABASE IF EXISTS system_meta")
        self._client.execute("CREATE DATABASE system_meta")
        self._client.execute("CREATE TABLE system_meta.databases (name String, engine String, comment String) engine = CnchMergeTree("
                             ") order by tuple()")
        databases_result = self._client.execute("SHOW DATABASES")
        databases_holder = []
        for database in databases_result:
            database_name = database[0]
            if database_name.startswith("dbt"):
                databases_holder.append(database)
        databases_result = databases_holder
        if len(databases_result) > 0:
            self._client.execute("INSERT INTO system_meta.databases VALUES", ((x[0], x[8], x[7]) for x in databases_result))

        if "system_meta.tables" not in sql and "system_meta.columns" not in sql:
            return

        self._client.execute("CREATE TABLE system_meta.tables (name String, database String, engine String, comment "
                             "String, type String) engine = CnchMergeTree() order by tuple()")

        table_names = []
        for database in databases_result:
            database_name = database[0]
            tables_result = self._client.execute(f"SHOW TABLES FROM {database_name}")
            for table in tables_result:
                table_name = table[0]
                table_names.append(f'{database_name}.{table_name}')
            if len(tables_result) > 0:
                self._client.execute("INSERT INTO system_meta.tables VALUES",
                                     ((x[0], database_name, "CnchMergeTree", x[7], x[8]) for x in tables_result))

        if "system_meta.columns" not in sql:
            return

        self._client.execute("CREATE TABLE system_meta.columns (database String, table String, name String, position "
                             "Int, type String, comment String) engine = CnchMergeTree() order by tuple()")
        for table_name in table_names:
            database_identifier = table_name.split(".")[0]
            table_identifier = table_name.split(".")[1]
            columns_result = self._client.execute(f"DESCRIBE TABLE {table_name}")
            # TODO: Verify ordering is the same as index
            columns_data = []
            idx = 0
            for x in columns_result:
                columns_data.append([database_identifier, table_identifier, x[0], idx, x[1], x[4]])
                idx = idx + 1
            if len(columns_result) > 0:
                self._client.execute("INSERT INTO system_meta.columns VALUES", columns_data)

    def rewrite_sql(self, sql):
        # TODO: Engine rewrite from upstream caller
        sql = sql.replace("MergeTree", "CnchMergeTree")
        # TODO: SETTINGS rewrite from upstream caller
        sql = sql.replace("SETTINGS  allow_nullable_key=1", "")
        return sql

    def get_rename_view_sql(self, sql):
        sql = self.modify_rename_table_syntax(sql)
        tokens = sql.split(" ")
        old_identifier = tokens[2]
        new_identifier = tokens[4]
        if not self.is_view(old_identifier):
            return sql
        old_view_create_table = self._client.execute(f'SHOW CREATE TABLE {old_identifier}')[0][0]
        tokens = old_view_create_table.split(" ")
        tokens[2] = new_identifier
        new_view_create_table = ' '.join(tokens)

        self._client.execute(f'DROP VIEW IF EXISTS {old_identifier}')
        return new_view_create_table

    def modify_drop_table_syntax(self, sql):
        tokens = sql.split(" ")
        table_identifier = ""
        for iter in range(0, len(tokens)-1):
            if tokens[iter].upper().endswith("DROP") and tokens[iter+1].upper() == "TABLE":
                table_identifier = tokens[iter+4]
                break
        if not self.is_view(table_identifier):
            return sql
        for iter in range(0, len(tokens)-1):
            if tokens[iter].upper().endswith("DROP") and tokens[iter+1].upper() == "TABLE":
                tokens[iter+1] = "view"
                break
        return ' '.join(tokens)

    def is_view(self, identifier):
        tokens = identifier.split(".")
        database_name = tokens[0]
        table_name = tokens[1]

        table_results = self._client.execute(f'SHOW TABLES FROM {database_name}')
        for table in table_results:
            name = table[0]
            table_type = table[8]
            if name == table_name and table_type == "VIEW":
                return True
        return False

    def modify_rename_table_syntax(self, sql):
        tokens = sql.split(" ")
        modify_sql = ""
        st = False
        for token in tokens:
            if token == "rename":
                st = True
            if st:
                modify_sql += token
                modify_sql += " "
        return modify_sql

    def handle_format_csv(self, sql):
        tokens = sql.split("|")

        insert_query = tokens[1]
        column_count = int(tokens[3])

        datatypes = []
        for i in range(4, 4 + column_count):
            datatypes.append(tokens[i])

        # TODO: Support all datatypes
        data = []
        for i in range(4 + column_count, len(tokens), column_count):
            row = []
            idx = 0
            for j in range(i, i + column_count):
                val = tokens[j]
                if datatypes[idx] == "String":
                    val = str(val)
                if datatypes[idx] == "Datetime":
                    val = parser.parse(val)
                if datatypes[idx] == "Int":
                    if str(val) == 'None':
                        val = 0
                    val = int(val)
                if datatypes[idx] == "BooleanInt":
                    cur = str(val).upper()
                    if cur == "TRUE":
                        val = 1
                    else:
                        val = 0
                row.append(val)
                idx += 1
            data.append(row)
        return insert_query, data


class NativeClientResult:
    def __init__(self, native_result):
        self.result_set = native_result[0]
        self.column_names = [col[0] for col in native_result[1]]


"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import re
import time
from contextlib import contextmanager
from typing import Any, Optional, Tuple

import agate
import dbt.exceptions
from dbt.adapters.sql import SQLConnectionManager
from dbt.contracts.connection import Connection
from dbt.events import AdapterLogger

from dbt.adapters.bytehouse.dbclient import BhRetryableException, get_db_client

logger = AdapterLogger('bytehouse')
retryable_exceptions = [BhRetryableException]
ddl_re = re.compile(r'^\s*(CREATE|DROP|ALTER)\s', re.IGNORECASE)


class ByteHouseConnectionManager(SQLConnectionManager):
    """
    ByteHouse Connector connection manager.
    """

    TYPE = 'bytehouse'

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except Exception as exp:
            logger.debug('Error running SQL: {}', sql)
            if isinstance(exp, dbt.exceptions.RuntimeException):
                raise
            raise dbt.exceptions.RuntimeException(exp) from exp

    @classmethod
    def open(cls, connection):
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection
        credentials = cls.get_credentials(connection.credentials)

        def connect():
            return get_db_client(credentials)

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=credentials.retries,
            retryable_exceptions=retryable_exceptions,
        )

    def cancel(self, connection):
        connection_name = connection.name
        logger.debug('Cancelling query \'{}\'', connection_name)
        connection.handle.close()
        logger.debug('Cancel query \'{}\'', connection_name)

    def release(self):
        pass  # There is no "release" type functionality in the existing ByteHouse connectors

    @classmethod
    def get_table_from_response(cls, response, column_names) -> agate.Table:
        """
        Build agate table from response.
        :param response: ByteHouse query result
        :param column_names: Table column names
        """
        data = []
        for row in response:
            data.append(dict(zip(column_names, row)))

        return dbt.clients.agate_helper.table_from_data_flat(data, column_names)

    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False
    ) -> Tuple[str, agate.Table]:
        # Don't try to fetch result of clustered DDL responses, we don't know what to do with them
        if fetch and ddl_re.match(sql):
            fetch = False

        sql = self._add_query_comment(sql)
        conn = self.get_thread_connection()
        client = conn.handle

        with self.exception_handler(sql):
            logger.debug(f'On {conn.name}: {sql}...')
            pre = time.time()
            if fetch:
                query_result = client.query(sql)
            else:
                query_result = client.command(sql)
            status = self.get_status(client)
            logger.debug(f'SQL status: {status} in {(time.time() - pre):.2f} seconds')
            if fetch:
                table = self.get_table_from_response(
                    query_result.result_set, query_result.column_names
                )
            else:
                table = dbt.clients.agate_helper.empty_table()
            return status, table

    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Optional[Any] = None,
        abridge_sql_log: bool = False,
    ) -> Tuple[Connection, Any]:
        sql = self._add_query_comment(sql)
        conn = self.get_thread_connection()
        client = conn.handle

        with self.exception_handler(sql):
            logger.debug(f'On {conn.name}: {sql}...')

            pre = time.time()
            client.command(sql)

            status = self.get_status(client)

            logger.debug(f'SQL status: {status} in {(time.time() - pre):0.2f} seconds')

            return conn, None

    @classmethod
    def get_credentials(cls, credentials):
        """
        Returns ByteHouse credentials
        """
        return credentials

    @classmethod
    def get_status(cls, _):
        """
        Returns connection status
        """
        return 'OK'

    @classmethod
    def get_response(cls, _):
        return 'OK'

    def begin(self):
        pass

    def commit(self):
        pass



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

#!/usr/bin/env python

import os
import re

from setuptools import find_namespace_packages, setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md')) as f:
    long_description = f.read()


# get this from a separate file
def _dbt_bytehouse_version():
    _version_path = os.path.join(this_directory, 'dbt', 'adapters', 'bytehouse', '__version__.py')
    _version_pattern = r'''version\s*=\s*["'](.+)["']'''
    with open(_version_path) as f:
        match = re.search(_version_pattern, f.read().strip())
        if match is None:
            raise ValueError(f'invalid version at {_version_path}')
        return match.group(1)


package_name = 'dbt-bytehouse'
package_version = _dbt_bytehouse_version()
description = '''The ByteHouse plugin for dbt (data build tool)'''

dbt_version = '1.3.0'
dbt_minor = '.'.join(dbt_version.split('.')[0:2])

if not package_version.startswith(dbt_minor):
    raise ValueError(
        f'Invalid setup.py: package_version={package_version} must start with '
        f'dbt_version={dbt_minor}'
    )

github_url = 'https://github.com/bytehouse-cloud/bytehouse-dbt'

setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rafsan Mazumder',
    author_email='rafsan.mazumder@bytedance.com',
    url=github_url,
    license='MIT',
    keywords='ByteHouse dbt connector',
    project_urls={
        'Documentation': github_url,
        'Changes': github_url + '/blob/main/CHANGELOG.md'
    },
    packages=find_namespace_packages(include=['dbt', 'dbt.*']),
    package_data={
        'dbt': [
            'include/bytehouse/dbt_project.yml',
            'include/bytehouse/macros/*.sql',
            'include/bytehouse/macros/**/*.sql',
        ]
    },
    install_requires=[
        f'dbt-core~={dbt_version}',
        'bytehouse-driver',
        'python-dateutil',
    ],
    python_requires=">=3.7",
    platforms='any',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



"""
   Copyright 2016-2022 ClickHouse, Inc.

   Copyright 2022- 2023 Bytedance Ltd. and/or its affiliates

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from dataclasses import dataclass
from typing import Optional

import dbt.exceptions
from dbt.adapters.base.relation import BaseRelation, Policy


@dataclass
class ByteHouseQuotePolicy(Policy):
    database: bool = False
    schema: bool = False
    identifier: bool = False


@dataclass
class ByteHouseIncludePolicy(Policy):
    database: bool = False
    schema: bool = True
    identifier: bool = True


@dataclass(frozen=True, eq=False, repr=False)
class ByteHouseRelation(BaseRelation):
    quote_policy: ByteHouseQuotePolicy = ByteHouseQuotePolicy()
    include_policy: ByteHouseIncludePolicy = ByteHouseIncludePolicy()
    quote_character: str = ''
    can_exchange: bool = False

    def __post_init__(self):
        if self.database != self.schema and self.database:
            raise dbt.exceptions.RuntimeException(
                f'Cannot set database {self.database} in bytehouse!'
            )

    def render(self):
        if self.include_policy.database and self.include_policy.schema:
            raise dbt.exceptions.RuntimeException(
                'Got a bytehouse relation with schema and database set to '
                'include, but only one can be set'
            )
        return super().render()

    def matches(
        self,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        identifier: Optional[str] = None,
    ):
        if schema:
            raise dbt.exceptions.RuntimeException(
                f'Passed unexpected schema value {schema} to Relation.matches'
            )
        return self.database == database and self.identifier == identifier
